#ifndef __UNIXLIB_LOADED
#define __UNIXLIB_LOADED	1

/*	UNIXLIB - V3.0 - Prototypes for UNIX emulation functions	*/

/* Changed for XPilot 3.0 */
/* we don't need any of the prototypes in the standard VMS unixlib */

#endif					/* __UNIXLIB_LOADED */
